'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var tslib = require('tslib');
var espression = require('espression');

/**
 * Copyright (c) 2018 Adrian Panella <ianchi74@outlook.com>
 *
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */
/* eslint-disable @typescript-eslint/ban-types */
/**
 * Represents the result of a JsonPath query.
 */
var JsonPathResult = /** @class */ (function () {
    function JsonPathResult(obj) {
        this.values = [];
        this.paths = [];
        if (obj instanceof JsonPathResult) {
            this.root = obj.root;
        }
        else {
            this.values.push(obj);
            this.paths.push(['$']);
            this.root = obj;
        }
    }
    // array like functions
    JsonPathResult.prototype.push = function (value, path) {
        this.values.push(value);
        this.paths.push(path);
    };
    JsonPathResult.prototype.slice = function (start, end, step, descendant) {
        var ret = new JsonPathResult(this);
        if (step === 0)
            return ret;
        var _step = toInt(step) || 1;
        var _start = toInt(start);
        var _end = toInt(end);
        this.forEach(function (val, path, _depth) {
            if (!Array.isArray(val))
                return;
            var first = _start === 0
                ? 0
                : _start
                    ? _start < 0
                        ? val.length + _start
                        : _start
                    : _step > 0
                        ? 0
                        : val.length - 1;
            if (first < 0 && _step > 0)
                first = 0;
            else if (_step < 0 && first >= val.length)
                first = val.length - 1;
            var last = _end === 0
                ? 0
                : _end
                    ? _end < 0
                        ? val.length + _end
                        : _end
                    : _step > 0
                        ? val.length
                        : -1;
            for (var i = first; ((_step > 0 && i < last) || (_step < 0 && i > last)) && i < val.length && i >= 0; i += _step) {
                ret.push(val[i], path.concat(i.toString()));
            }
        }, descendant ? undefined : 0);
        return ret;
    };
    JsonPathResult.prototype.forEach = function (callback, maxDepth) {
        for (var i = 0; i < this.values.length; i++) {
            if (callback)
                callback(this.values[i], this.paths[i], 0);
            if (maxDepth !== 0)
                traverse(this.values[i], this.paths[i], maxDepth, callback, 1);
        }
    };
    JsonPathResult.prototype.concat = function (partial) {
        var ret = new JsonPathResult(this);
        if (partial.root !== this.root)
            return null;
        ret.values = this.values.concat(partial.values);
        ret.paths = this.paths.concat(partial.paths);
        return ret;
    };
    return JsonPathResult;
}());
// auxiliary functions
function traverse(obj, root, maxDepth, callback, depth) {
    var path;
    if (Array.isArray(obj)) {
        for (var i = 0; i < obj.length; i++) {
            path = root.concat(i.toString());
            if (callback)
                callback(obj[i], path, depth);
            if (!maxDepth || depth < maxDepth)
                traverse(obj[i], path, maxDepth, callback, depth + 1);
        }
    }
    else if (typeof obj === 'object') {
        for (var prop in obj) {
            path = root.concat(prop);
            if (callback)
                callback(obj[prop], path, depth);
            if (!maxDepth || depth < maxDepth)
                traverse(obj[prop], path, maxDepth, callback, depth + 1);
        }
    }
}
function toInt(val) {
    var ret;
    if (typeof val === 'string') {
        if (!val) {
            ret = undefined;
        }
        else {
            ret = parseInt(val, 10);
            if (Number.isNaN(ret))
                throw new Error('Invalid number');
        }
    }
    else
        ret = val;
    return ret;
}

/* eslint-disable max-classes-per-file */
var JPCHILD_EXP = 'JPChildExpression';
var JPDESC_EXP = 'JPDescendantExpression';
var JPUNION_EXP = 'JPUnionExpression';
var JPFILTER_EXP = 'JPFilterExpression';
var JPSLICE_EXP = 'JPSliceExpression';
var JPWILDCARD_EXP = 'JPWildcard';
var JPEXP_EXP = 'JPExpression';
var COMPUTED_CONF = {
    type: JPCHILD_EXP,
    left: espression.OBJECT,
    right: espression.PROPERTY,
    extra: { computed: true },
    subRules: 'computed',
    close: ']',
};
var MEMBER_CONF = {
    type: JPCHILD_EXP,
    left: espression.OBJECT,
    right: espression.PROPERTY,
    extra: { computed: false },
    subRules: JPCHILD_EXP,
};
function jsonPathRules(identStart, identPart) {
    // properties can have reserved words as names
    var wildcardRule = new espression.IdentifierRule({
        typeIdent: JPWILDCARD_EXP,
        identStart: { re: /[*]/, re2: undefined },
        // eslint-disable-next-line no-empty-character-class
        identPart: { re: /[]/, re2: undefined },
    });
    // adds '@' as valid identifier start
    var rules = espression.esNextRules(tslib.__assign(tslib.__assign({}, identStart), { re: /[@$_A-Za-z]/ }), identPart);
    rules[JPEXP_EXP] = [
        new espression.BinaryOperatorRule({
            '.': MEMBER_CONF,
            '[': COMPUTED_CONF,
            '..': tslib.__assign(tslib.__assign({}, MEMBER_CONF), { type: JPDESC_EXP }),
            '..[': tslib.__assign(tslib.__assign({}, COMPUTED_CONF), { type: JPDESC_EXP }),
        }),
        new espression.UnaryOperatorRule({ '(': { close: ')', subRules: espression.EXPRESSION, type: 'JProotExpr' } }),
        new espression.IdentifierRule({ typeIdent: 'JPRoot' }),
    ];
    rules[JPCHILD_EXP] = [new espression.IdentifierRule(), wildcardRule];
    rules.computed = [
        new espression.MultiOperatorRule({
            type: JPUNION_EXP,
            prop: espression.EXPRESSIONS,
            separators: ',',
        }),
        new espression.UnaryOperatorRule({
            '(': {
                type: JPEXP_EXP,
                prop: espression.EXPRESSION,
                close: ')',
                subRules: espression.NOCOMMA_EXPR,
            },
        }),
        new espression.UnaryOperatorRule({
            '?(': {
                type: JPFILTER_EXP,
                prop: espression.EXPRESSION,
                close: ')',
                subRules: espression.NOCOMMA_EXPR,
            },
        }),
        wildcardRule,
        new espression.MultiOperatorRule({
            type: JPSLICE_EXP,
            prop: espression.EXPRESSIONS,
            separators: ':',
            sparse: true,
            trailling: true,
            maxSep: 2,
            types: [espression.LITERAL_EXP, espression.UNARY_EXP],
        }),
        new espression.UnaryOperatorRule(espression.opConf(['-', '+'], espression.UNARY_TYPE_PRE)),
        new espression.NumberRule({ radix: 10, decimal: false }),
        new espression.StringRule(),
    ];
    return rules;
}
/**
 * Extended ES-next parser.
 *  add the '<$..path>' notation as jsonPath literal, with priority as first token
 *
 */
var ESPathParser = /** @class */ (function (_super) {
    tslib.__extends(ESPathParser, _super);
    function ESPathParser() {
        var _this = this;
        var esPathRules = jsonPathRules();
        esPathRules[espression.TOKEN].unshift(new espression.UnaryOperatorRule({ '<': { close: '>', subRules: JPEXP_EXP } }));
        _this = _super.call(this, esPathRules, espression.STATEMENT) || this;
        return _this;
    }
    return ESPathParser;
}(espression.Parser));
var JsonPathParser = /** @class */ (function (_super) {
    tslib.__extends(JsonPathParser, _super);
    function JsonPathParser() {
        return _super.call(this, jsonPathRules(), JPEXP_EXP) || this;
    }
    return JsonPathParser;
}(espression.Parser));

/**
 * Copyright (c) 2018 Adrian Panella <ianchi74@outlook.com>
 *
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */
/**
 * Static-eval extension to evaluate JsonPath AST nodes
 */
var JsonPathStaticEval = /** @class */ (function (_super) {
    tslib.__extends(JsonPathStaticEval, _super);
    function JsonPathStaticEval() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    JsonPathStaticEval.prototype.JPRoot = function (node, context) {
        return new JsonPathResult(context[node.name]);
    };
    JsonPathStaticEval.prototype.JPRootExpr = function (node, context) {
        return new JsonPathResult(this._eval(node.argument, context));
    };
    JsonPathStaticEval.prototype.JPChildExpression = function (node, context) {
        return this.evalMember(this._eval(node.object, context), node, false, context);
    };
    JsonPathStaticEval.prototype.JPDescendantExpression = function (node, context) {
        return this.evalMember(this._eval(node.object, context), node, true, context);
    };
    JsonPathStaticEval.prototype.evalMember = function (obj, node, descendant, context) {
        var _this = this;
        var props = node.property.type === JPUNION_EXP ? node.property.expressions : [node.property];
        var childContext = Object.create(context);
        return props.reduce(function (acum, n) {
            var ret = new JsonPathResult(obj);
            var member;
            switch (n.type) {
                case JPEXP_EXP:
                    obj.forEach(function (val, path, _depth) {
                        if (typeof val === 'object') {
                            childContext['@'] = val;
                            member = _this._eval(n.expression, childContext);
                            if (member in val)
                                ret.push(val[member], path.concat(member));
                        }
                    }, descendant ? undefined : 0);
                    break;
                case JPFILTER_EXP:
                    obj.forEach(function (val, path, depth) {
                        if (!depth)
                            return; // filter applies on children
                        childContext['@'] = val;
                        if (_this._eval(n.expression, childContext))
                            ret.push(val, path);
                    }, descendant ? undefined : 1);
                    break;
                case JPSLICE_EXP:
                    var idx = n.expressions.map(function (i) { return i && _this._eval(i, childContext); });
                    ret = obj.slice(idx[0], idx[1], idx[2], descendant);
                    break;
                case JPWILDCARD_EXP:
                    obj.forEach(function (val, path, depth) {
                        if (!depth)
                            return; // applies on children
                        ret.push(val, path);
                    }, descendant ? undefined : 1);
                    break;
                default:
                    member = node.computed ? _this._eval(n, context) : n.name;
                    obj.forEach(function (val, path, _depth) {
                        if (Array.isArray(val) &&
                            typeof member === 'number' &&
                            member < 0 &&
                            val.length + member in val) {
                            ret.push(val[val.length + member], path.concat((val.length + member).toString()));
                        }
                        if (typeof val === 'object' && member in val) {
                            ret.push(val[member], path.concat(member));
                        }
                    }, descendant ? undefined : 0);
                    break;
            }
            return acum ? acum.concat(ret) : ret;
        }, null);
    };
    return JsonPathStaticEval;
}(espression.ES6StaticEval));

/*!
 * Copyright (c) 2018 Adrian Panella <ianchi74@outlook.com>
 *
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */
var JsonPath = /** @class */ (function () {
    function JsonPath() {
    }
    JsonPath.prototype.evaluate = function (ast, ctx) {
        return JsonPath.jpEval.evaluate(ast, ctx);
    };
    JsonPath.prototype.parse = function (path) {
        return JsonPath.espParser.parse(path);
    };
    JsonPath.prototype.eval = function (path, ctx) {
        return JsonPath.jpEval.evaluate(JsonPath.espParser.parse(path), ctx);
    };
    // eslint-disable-next-line @typescript-eslint/ban-types
    JsonPath.prototype.query = function (obj, path) {
        return JsonPath.jpEval.evaluate(JsonPath.jpParser.parse(path), { $: obj });
    };
    JsonPath.jpEval = new JsonPathStaticEval();
    JsonPath.jpParser = new JsonPathParser();
    JsonPath.espParser = new ESPathParser();
    return JsonPath;
}());

exports.ESPathParser = ESPathParser;
exports.JsonPath = JsonPath;
exports.JsonPathParser = JsonPathParser;
exports.JsonPathResult = JsonPathResult;
exports.JsonPathStaticEval = JsonPathStaticEval;
//# sourceMappingURL=espression-jsonpath.cjs.js.map
