import { Parser, TOKEN, UnaryOperatorRule, STATEMENT, IdentifierRule, esNextRules, BinaryOperatorRule, EXPRESSION, MultiOperatorRule, EXPRESSIONS, NOCOMMA_EXPR, LITERAL_EXP, UNARY_EXP, opConf, UNARY_TYPE_PRE, NumberRule, StringRule, OBJECT, PROPERTY, ES6StaticEval } from 'espression';

/**
 * Copyright (c) 2018 Adrian Panella <ianchi74@outlook.com>
 *
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */
/* eslint-disable @typescript-eslint/ban-types */
/**
 * Represents the result of a JsonPath query.
 */
class JsonPathResult {
    constructor(obj) {
        this.values = [];
        this.paths = [];
        if (obj instanceof JsonPathResult) {
            this.root = obj.root;
        }
        else {
            this.values.push(obj);
            this.paths.push(['$']);
            this.root = obj;
        }
    }
    // array like functions
    push(value, path) {
        this.values.push(value);
        this.paths.push(path);
    }
    slice(start, end, step, descendant) {
        const ret = new JsonPathResult(this);
        if (step === 0)
            return ret;
        const _step = toInt(step) || 1;
        const _start = toInt(start);
        const _end = toInt(end);
        this.forEach((val, path, _depth) => {
            if (!Array.isArray(val))
                return;
            let first = _start === 0
                ? 0
                : _start
                    ? _start < 0
                        ? val.length + _start
                        : _start
                    : _step > 0
                        ? 0
                        : val.length - 1;
            if (first < 0 && _step > 0)
                first = 0;
            else if (_step < 0 && first >= val.length)
                first = val.length - 1;
            const last = _end === 0
                ? 0
                : _end
                    ? _end < 0
                        ? val.length + _end
                        : _end
                    : _step > 0
                        ? val.length
                        : -1;
            for (let i = first; ((_step > 0 && i < last) || (_step < 0 && i > last)) && i < val.length && i >= 0; i += _step) {
                ret.push(val[i], path.concat(i.toString()));
            }
        }, descendant ? undefined : 0);
        return ret;
    }
    forEach(callback, maxDepth) {
        for (let i = 0; i < this.values.length; i++) {
            if (callback)
                callback(this.values[i], this.paths[i], 0);
            if (maxDepth !== 0)
                traverse(this.values[i], this.paths[i], maxDepth, callback, 1);
        }
    }
    concat(partial) {
        const ret = new JsonPathResult(this);
        if (partial.root !== this.root)
            return null;
        ret.values = this.values.concat(partial.values);
        ret.paths = this.paths.concat(partial.paths);
        return ret;
    }
}
// auxiliary functions
function traverse(obj, root, maxDepth, callback, depth) {
    let path;
    if (Array.isArray(obj)) {
        for (let i = 0; i < obj.length; i++) {
            path = root.concat(i.toString());
            if (callback)
                callback(obj[i], path, depth);
            if (!maxDepth || depth < maxDepth)
                traverse(obj[i], path, maxDepth, callback, depth + 1);
        }
    }
    else if (typeof obj === 'object') {
        for (const prop in obj) {
            path = root.concat(prop);
            if (callback)
                callback(obj[prop], path, depth);
            if (!maxDepth || depth < maxDepth)
                traverse(obj[prop], path, maxDepth, callback, depth + 1);
        }
    }
}
function toInt(val) {
    let ret;
    if (typeof val === 'string') {
        if (!val) {
            ret = undefined;
        }
        else {
            ret = parseInt(val, 10);
            if (Number.isNaN(ret))
                throw new Error('Invalid number');
        }
    }
    else
        ret = val;
    return ret;
}

/* eslint-disable max-classes-per-file */
const JPCHILD_EXP = 'JPChildExpression';
const JPDESC_EXP = 'JPDescendantExpression';
const JPUNION_EXP = 'JPUnionExpression';
const JPFILTER_EXP = 'JPFilterExpression';
const JPSLICE_EXP = 'JPSliceExpression';
const JPWILDCARD_EXP = 'JPWildcard';
const JPEXP_EXP = 'JPExpression';
const COMPUTED_CONF = {
    type: JPCHILD_EXP,
    left: OBJECT,
    right: PROPERTY,
    extra: { computed: true },
    subRules: 'computed',
    close: ']',
};
const MEMBER_CONF = {
    type: JPCHILD_EXP,
    left: OBJECT,
    right: PROPERTY,
    extra: { computed: false },
    subRules: JPCHILD_EXP,
};
function jsonPathRules(identStart, identPart) {
    // properties can have reserved words as names
    const wildcardRule = new IdentifierRule({
        typeIdent: JPWILDCARD_EXP,
        identStart: { re: /[*]/, re2: undefined },
        // eslint-disable-next-line no-empty-character-class
        identPart: { re: /[]/, re2: undefined },
    });
    // adds '@' as valid identifier start
    const rules = esNextRules({ ...identStart, re: /[@$_A-Za-z]/ }, identPart);
    rules[JPEXP_EXP] = [
        new BinaryOperatorRule({
            '.': MEMBER_CONF,
            '[': COMPUTED_CONF,
            '..': { ...MEMBER_CONF, type: JPDESC_EXP },
            '..[': { ...COMPUTED_CONF, type: JPDESC_EXP },
        }),
        new UnaryOperatorRule({ '(': { close: ')', subRules: EXPRESSION, type: 'JProotExpr' } }),
        new IdentifierRule({ typeIdent: 'JPRoot' }),
    ];
    rules[JPCHILD_EXP] = [new IdentifierRule(), wildcardRule];
    rules.computed = [
        new MultiOperatorRule({
            type: JPUNION_EXP,
            prop: EXPRESSIONS,
            separators: ',',
        }),
        new UnaryOperatorRule({
            '(': {
                type: JPEXP_EXP,
                prop: EXPRESSION,
                close: ')',
                subRules: NOCOMMA_EXPR,
            },
        }),
        new UnaryOperatorRule({
            '?(': {
                type: JPFILTER_EXP,
                prop: EXPRESSION,
                close: ')',
                subRules: NOCOMMA_EXPR,
            },
        }),
        wildcardRule,
        new MultiOperatorRule({
            type: JPSLICE_EXP,
            prop: EXPRESSIONS,
            separators: ':',
            sparse: true,
            trailling: true,
            maxSep: 2,
            types: [LITERAL_EXP, UNARY_EXP],
        }),
        new UnaryOperatorRule(opConf(['-', '+'], UNARY_TYPE_PRE)),
        new NumberRule({ radix: 10, decimal: false }),
        new StringRule(),
    ];
    return rules;
}
/**
 * Extended ES-next parser.
 *  add the '<$..path>' notation as jsonPath literal, with priority as first token
 *
 */
class ESPathParser extends Parser {
    constructor() {
        const esPathRules = jsonPathRules();
        esPathRules[TOKEN].unshift(new UnaryOperatorRule({ '<': { close: '>', subRules: JPEXP_EXP } }));
        super(esPathRules, STATEMENT);
    }
}
class JsonPathParser extends Parser {
    constructor() {
        super(jsonPathRules(), JPEXP_EXP);
    }
}

/**
 * Copyright (c) 2018 Adrian Panella <ianchi74@outlook.com>
 *
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */
/**
 * Static-eval extension to evaluate JsonPath AST nodes
 */
class JsonPathStaticEval extends ES6StaticEval {
    JPRoot(node, context) {
        return new JsonPathResult(context[node.name]);
    }
    JPRootExpr(node, context) {
        return new JsonPathResult(this._eval(node.argument, context));
    }
    JPChildExpression(node, context) {
        return this.evalMember(this._eval(node.object, context), node, false, context);
    }
    JPDescendantExpression(node, context) {
        return this.evalMember(this._eval(node.object, context), node, true, context);
    }
    evalMember(obj, node, descendant, context) {
        const props = node.property.type === JPUNION_EXP ? node.property.expressions : [node.property];
        const childContext = Object.create(context);
        return props.reduce((acum, n) => {
            let ret = new JsonPathResult(obj);
            let member;
            switch (n.type) {
                case JPEXP_EXP:
                    obj.forEach((val, path, _depth) => {
                        if (typeof val === 'object') {
                            childContext['@'] = val;
                            member = this._eval(n.expression, childContext);
                            if (member in val)
                                ret.push(val[member], path.concat(member));
                        }
                    }, descendant ? undefined : 0);
                    break;
                case JPFILTER_EXP:
                    obj.forEach((val, path, depth) => {
                        if (!depth)
                            return; // filter applies on children
                        childContext['@'] = val;
                        if (this._eval(n.expression, childContext))
                            ret.push(val, path);
                    }, descendant ? undefined : 1);
                    break;
                case JPSLICE_EXP:
                    const idx = n.expressions.map((i) => i && this._eval(i, childContext));
                    ret = obj.slice(idx[0], idx[1], idx[2], descendant);
                    break;
                case JPWILDCARD_EXP:
                    obj.forEach((val, path, depth) => {
                        if (!depth)
                            return; // applies on children
                        ret.push(val, path);
                    }, descendant ? undefined : 1);
                    break;
                default:
                    member = node.computed ? this._eval(n, context) : n.name;
                    obj.forEach((val, path, _depth) => {
                        if (Array.isArray(val) &&
                            typeof member === 'number' &&
                            member < 0 &&
                            val.length + member in val) {
                            ret.push(val[val.length + member], path.concat((val.length + member).toString()));
                        }
                        if (typeof val === 'object' && member in val) {
                            ret.push(val[member], path.concat(member));
                        }
                    }, descendant ? undefined : 0);
                    break;
            }
            return acum ? acum.concat(ret) : ret;
        }, null);
    }
}

/*!
 * Copyright (c) 2018 Adrian Panella <ianchi74@outlook.com>
 *
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */
class JsonPath {
    evaluate(ast, ctx) {
        return JsonPath.jpEval.evaluate(ast, ctx);
    }
    parse(path) {
        return JsonPath.espParser.parse(path);
    }
    eval(path, ctx) {
        return JsonPath.jpEval.evaluate(JsonPath.espParser.parse(path), ctx);
    }
    // eslint-disable-next-line @typescript-eslint/ban-types
    query(obj, path) {
        return JsonPath.jpEval.evaluate(JsonPath.jpParser.parse(path), { $: obj });
    }
}
JsonPath.jpEval = new JsonPathStaticEval();
JsonPath.jpParser = new JsonPathParser();
JsonPath.espParser = new ESPathParser();

export { ESPathParser, JsonPath, JsonPathParser, JsonPathResult, JsonPathStaticEval };
//# sourceMappingURL=espression-jsonpath.js.map
